import logging
from datetime import datetime, timezone
from platform import python_version
from typing import Dict, Optional, List, Callable, cast, Any

from frogml_core.clients.jfrog_gateway import JfrogGatewayClient
from frogml_core.clients.model_version_manager import ModelVersionManagerClient
from frogml_core.exceptions import FrogmlException
from frogml_core.utils.proto_utils import ProtoUtils
from frogml_proto.jfml.model_version.v1.model_version_framework_pb2 import (
    ModelVersionFramework,
)
from frogml_proto.qwak.jfrog.gateway.v0.repository_service_pb2 import (
    GetRepositoryConfigurationResponse,
)
from frogml_storage.frog_ml import SerializationMetadata, FrogMLStorage
from frogml_storage.model_manifest import ModelManifest
from frogml_storage.models.frogml_model_version import FrogMLModelVersion

from frogml.finders.model_class_info import ModelClassInfo
from frogml.finders.model_finder import ModelFinder
from frogml.modifiers.model_modifier import ModelModifier
from frogml.utils.dependencies_tools import _dependency_files_handler
from frogml.utils.files_tools import (
    _zip_dir,
    _copy_dir_without_ignored_files,
    _remove_dir,
)
from frogml.utils.validations import _validate_load_model

_STORAGE_MODEL_ENTITY_TYPE = "model"
_PYTHON_RUNTIME = "python"

_logger = logging.getLogger(__name__)


def _get_model_metadata(
    model_flavor: str,
    model_flavor_version: str,
    serialization_format: str,
) -> SerializationMetadata:
    return SerializationMetadata(
        framework=model_flavor,
        framework_version=model_flavor_version,
        serialization_format=serialization_format,
        runtime=_PYTHON_RUNTIME,
        runtime_version=python_version(),
    )


def _get_model_info_from_artifactory(
    repository: str,
    model_name: str,
    version: str,
) -> Dict:
    return FrogMLStorage().get_entity_manifest(
        entity_type=_STORAGE_MODEL_ENTITY_TYPE,
        repository=repository,
        entity_name=model_name,
        version=version,
        namespace=None,
    )


def _download_model_version_from_artifactory(
    model_flavor: str,
    repository: str,
    model_name: str,
    version: str,
    model_framework: str,
    download_target_path: str,
    deserializer: Callable,
):
    """
    Download model version from artifactory
    :param model_flavor: model flavor files/catboost etc.
    :param repository: repository name
    :param model_name: the name of the model
    :param version: version of the model
    :param download_target_path: the path to download the model to
    :param model_framework: model framework files/catboost etc.
    :return: Loaded model
    """
    _validate_load_model(
        repository=repository,
        model_name=model_name,
        version=version,
        model_framework=model_framework,
        model_flavor=model_flavor,
    )

    FrogMLStorage().download_model_version(
        repository=repository,
        model_name=model_name,
        version=version,
        target_path=download_target_path,
    )
    return deserializer()


def _log_model(
    model_name: str,
    target_dir: str,
    model_flavor: str,
    framework_version: str,
    full_model_path: str,
    serialization_format: str,
    repository: str,
    version: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    dependencies: Optional[List[str]] = None,
    code_dir: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    metrics: Optional[Dict[str, Any]] = None,
) -> None:
    project_key: str = __get_project_key(repository)
    model_version_framework: ModelVersionFramework = (
        ProtoUtils.model_framework_from_file_format(
            serialization_format, framework_version
        )
    )

    if not version:
        version = datetime.now(timezone.utc).strftime("%Y-%m-%d-%H-%M-%S-%f")[:-3]

    __validate_model_version(
        project_key, repository, model_name, version, model_version_framework
    )

    properties: Dict[str, str] = merge_properties_parameters_metrics(
        metrics, parameters, properties
    )

    dependencies = _dependency_files_handler(
        dependencies=dependencies, target_dir=target_dir
    )
    metadata: SerializationMetadata = _get_model_metadata(
        model_flavor=model_flavor,
        model_flavor_version=framework_version,
        serialization_format=serialization_format,
    )

    zipped_code_path: Optional[str] = build_zip_code_path(
        code_dir, target_dir, model_flavor, model_name, repository, version
    )

    frog_ml_model_version: FrogMLModelVersion = __log_model_to_artifactory(
        dependencies,
        full_model_path,
        model_name,
        properties,
        repository,
        version,
        metadata,
        zipped_code_path,
    )

    manifest: ModelManifest = cast(ModelManifest, frog_ml_model_version.entity_manifest)

    __create_model_version(
        project_key,
        repository,
        model_name,
        version,
        model_version_framework,
        manifest,
        parameters,
        metrics,
    )


def build_zip_code_path(
    code_dir_path: Optional[str],
    parent_dir_path: str,
    model_flavor: str,
    model_name: str,
    repository: str,
    version: str,
) -> Optional[str]:
    """
    Build the path to the zipped code directory.
    Before zipping, the code directory is copied to a target directory and filtered to remove ignored files.
    Then, the code directory is processed to find and modify model classes.
    The resulting code directory is zipped and the path to the zipped code directory is returned.

    :param code_dir_path: The path to the code directory
    :param model_flavor: The model flavor
    :param model_name: The model name
    :param repository: The repository key
    :param parent_dir_path: The parent directory of the zip file
    :param version: The model version
    :return: The path to the zipped code directory
    """
    zipped_code_path: Optional[str] = None

    if code_dir_path is None:
        return zipped_code_path

    filtered_copied_dir_path: str = _copy_dir_without_ignored_files(
        code_dir_path, parent_dir_path
    )
    try:
        process_code_directory(
            filtered_copied_dir_path, repository, model_name, version, model_flavor
        )
        zipped_code_path = _zip_dir(
            root_dir=parent_dir_path, dir_to_zip=filtered_copied_dir_path
        )
    except Exception as e:
        raise FrogmlException(f"Failed to zip code directory: {e}") from e
    finally:
        _remove_dir(filtered_copied_dir_path)

    return zipped_code_path


def merge_properties_parameters_metrics(
    metrics: Optional[Dict[str, Any]] = None,
    parameters: Optional[Dict[str, Any]] = None,
    properties: Optional[Dict[str, str]] = None,
) -> Dict[str, str]:
    result = {}
    if properties is not None:
        result.update(properties)

    if parameters is not None:
        result.update(parameters)

    if metrics is not None:
        result.update(metrics)

    result = {key: str(value) for key, value in result.items()}

    return result


def __get_project_key(repository_key: str) -> str:
    _logger.info("Getting project key for repository %s", repository_key)

    jfrog_gateway_client = JfrogGatewayClient()
    response: GetRepositoryConfigurationResponse = (
        jfrog_gateway_client.get_repository_configuration(repository_key=repository_key)
    )

    if not response.repository_spec or not response.repository_spec.project_key:
        raise ValueError(
            f"Repository '{repository_key}' does not belong to any project"
        )

    return response.repository_spec.project_key


def __create_model_version(
    project_key: str,
    repository: str,
    model_name: str,
    model_version: str,
    model_version_framework: ModelVersionFramework,
    manifest: ModelManifest,
    parameters: Optional[Dict[str, Any]] = None,
    metrics: Optional[Dict[str, Any]] = None,
):
    model_version_manager_client = ModelVersionManagerClient()

    try:
        model_version_manager_client.create_model_version(
            project_key=project_key,
            repository_key=repository,
            model_name=model_name,
            model_version_name=model_version,
            model_version_framework=model_version_framework,
            model_artifact=manifest.artifacts,
            dependency_artifacts=manifest.dependency_artifacts,
            code_artifacts=(
                [manifest.code_artifacts] if manifest.code_artifacts else None
            ),
            parameters=parameters,
            metrics=metrics,
        )
    except Exception as e:
        _logger.warning(
            "Failed to create model version %s in JFML due to this following error: %s\n"
            "Before retrying, please delete the uploaded artifact from Artifactory",
            model_version,
            e,
        )


def __validate_model_version(
    project_key: str,
    repository: str,
    model_name: str,
    model_version: str,
    model_version_framework: ModelVersionFramework,
):
    try:
        model_version_manager_client = ModelVersionManagerClient()
        model_version_manager_client.validate_create_model_version(
            project_key=project_key,
            repository_key=repository,
            model_name=model_name,
            model_version_name=model_version,
            model_version_framework=model_version_framework,
            model_artifact=[],
        )
    except Exception as e:
        raise ValueError(str(e)) from e


def __log_model_to_artifactory(
    dependencies: Optional[List[str]],
    full_model_path: str,
    model_name: str,
    properties: Optional[Dict[str, str]],
    repository: str,
    version: str,
    metadata: SerializationMetadata,
    code_archive_file_path: Optional[str],
) -> FrogMLModelVersion:
    return FrogMLStorage().upload_model_version(
        repository=repository,
        model_name=model_name,
        model_path=full_model_path,
        model_type=metadata,
        version=version,
        properties=properties,
        dependencies_files_paths=dependencies,
        code_archive_file_path=code_archive_file_path,
    )


def process_code_directory(
    code_dir: str, repository: str, model_name: str, version: str, framework_name: str
):
    """
    Process a directory of Python files to find and modify model classes.

    :param code_dir: Directory containing Python files to analyze
    :param repository: Repository key to use in the template
    :param model_name: Model name to use in the template
    :param version: Model version to use in the template
    :param framework_name: The name of the framework used to train the model

    :raises ValueError: If no valid model class is found or if the model class doesn't have a model field
    """
    model_class_info: ModelClassInfo = ModelFinder.find_model_class(code_dir)

    if model_class_info is None:
        raise ValueError(
            "No class inheriting from FrogMlModel found in the code directory"
        )

    if model_class_info.model_field_name is None:
        raise ValueError(
            f"The class {model_class_info.class_name} does not have a model field (model, _model, or __model)"
        )

    ModelModifier.modify_model_file(
        model_class_info,
        repository,
        model_name,
        version,
        framework_name,
    )

    print(
        f"Successfully modified {model_class_info.file_path} for class {model_class_info.class_name}"
    )
